# -*- coding: utf-8 -*-

#   Copyright 2023 Brooks Su
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Collects some common icons in unicode character set.
"""

from enum import StrEnum


class UnicodeIcon(StrEnum):
    """Collects some common icons in unicode character set.
    """
    CAFE = '\u2615'
    WHITE_CIRCLE = '\u26aa'
    BLACK_CIRCLE = '\u26ab'
    FOOTBALL = '\u26bd'
    STOP = '\u26d4'
    GREEN_TICK = '\u2705'
    SMALL_WHITE_STAR = '\u2729'
    SMALL_PINWHEEL_STAR = '\u272f'
    SNOW_FLAKE = '\u2744'
    RED_CROSS = '\u274c'
    GREEN_CROSS = '\u274e'
    BLACK_SQUARE = '\u2b1b'
    WHITE_SQUARE = '\u2b1c'
    GOLD_STAR = '\u2b50'
    RED_RING = '\u2b55'

    SOS = '\U0001f198'
    RAINBOW = '\U0001f308'
    DARK_FACE = '\U0001f31a'
    GOLD_FACE = '\U0001f31d'
    TULIP = '\U0001f337'
    ROSE = '\U0001f339'
    FLOWER = '\U0001f33c'
    CORN = '\U0001f33d'
    MUSHROOM = '\U0001f344'
    TOMATO = '\U0001f345'
    RED_APPLE = '\U0001f34e'
    GREEN_APPLE = '\U0001f34f'
    PEAR = '\U0001f350'
    PEACH = '\U0001f351'
    CHERRY = '\U0001f352'
    HAM = '\U0001f356'
    GIFT = '\U0001f381'
    PUMPKIN = '\U0001f383'
    SAI = '\U0001f3ad'
    DICE = '\U0001f3b2'
    BASKETBALL = '\U0001f3c0'
    VOLLEYBALL = '\U0001f3d0'
    MICE = '\U0001f401'
    OCTOPUS = '\U0001f419'
    TIGER = '\U0001f42f'
    FROG = '\U0001f438'
    BEAR = '\U0001f43b'
    PANDA = '\U0001f43c'
    CROWN = '\U0001f451'
    JELLYFISH = '\U0001f47b'
    ALIEN = '\U0001f47d'
    SKULL = '\U0001f480'
    KEY = '\U0001f511'
    LOCK = '\U0001f512'
    RED_CIRCLE = '\U0001f534'
    BLUE_CIRCLE = '\U0001f535'
    ORANGE_DIAMOND = '\U0001f536'
    BLUE_DIAMOND = '\U0001f537'
    SMALL_ORANGE_DIAMOND = '\U0001f538'
    SMALL_BLUE_DIAMOND = '\U0001f539'
    ORANGE_CIRCLE = '\U0001f7e0'
    YELLOW_CIRCLE = '\U0001f7e1'
    GREEN_CIRCLE = '\U0001f7e2'
    MAGENTA_CIRCLE = '\U0001f7e3'
    BROWN_CIRCLE = '\U0001f7e4'
    RED_SQUARE = '\U0001f7e5'
    BLUE_SQUARE = '\U0001f7e6'
    ORANGE_SQUARE = '\U0001f7e7'
    YELLOW_SQUARE = '\U0001f7e8'
    GREEN_SQUARE = '\U0001f7e9'
    MAGENTA_SQUARE = '\U0001f7ea'
    BROWN_SQUARE = '\U0001f7eb'
    CRAB = '\U0001f980'
    BUTTERFLY = '\U0001f98b'
    SWAN = '\U0001f9a2'
